@@ -2,7 +2,7 @@
 mod support;
 
 macro_rules! impl_mat3_tests {
-    ($t:ident, $const_new:ident, $newmat3:ident, $mat3:ident, $mat4:ident, $newvec3:ident, $vec3:ident, $vec2:ident) => {
+    ($t:ident, $const_new:ident, $newmat3:ident, $mat3:ident, $mat2:ident, $mat4:ident, $quat:ident, $newvec3:ident, $vec3:ident, $vec2:ident) => {
         use core::$t::INFINITY;
         use core::$t::NAN;
         use core::$t::NEG_INFINITY;
@@ -70,6 +70,20 @@ macro_rules! impl_mat3_tests {
             assert_eq!($newvec3(1.0, 4.0, 7.0), m.row(0));
             assert_eq!($newvec3(2.0, 5.0, 8.0), m.row(1));
             assert_eq!($newvec3(3.0, 6.0, 9.0), m.row(2));
+
+            *m.col_mut(0) = m.col(0).zyx();
+            *m.col_mut(1) = m.col(1).zyx();
+            *m.col_mut(2) = m.col(2).zyx();
+            assert_eq!($newvec3(3.0, 2.0, 1.0), m.col(0));
+            assert_eq!($newvec3(6.0, 5.0, 4.0), m.col(1));
+            assert_eq!($newvec3(9.0, 8.0, 7.0), m.col(2));
+
+            should_panic!({ $mat3::ZERO.col(3) });
+            should_panic!({
+                let mut m = $mat3::ZERO;
+                m.col_mut(3);
+            });
+            should_panic!({ $mat3::ZERO.row(3) });
         }
 
         #[test]
@@ -98,12 +112,23 @@ macro_rules! impl_mat3_tests {
             let rot_x1 = $mat3::from_rotation_x(deg(180.0));
             let rot_x2 = $mat3::from_axis_angle($vec3::X, deg(180.0));
             assert_approx_eq!(rot_x1, rot_x2);
+            let rot_x3 = $mat3::from_quat($quat::from_rotation_x(deg(180.0)));
+            assert_approx_eq!(rot_x1, rot_x3);
+
             let rot_y1 = $mat3::from_rotation_y(deg(180.0));
             let rot_y2 = $mat3::from_axis_angle($vec3::Y, deg(180.0));
             assert_approx_eq!(rot_y1, rot_y2);
+            let rot_y3 = $mat3::from_quat($quat::from_rotation_y(deg(180.0)));
+            assert_approx_eq!(rot_y1, rot_y3);
+
             let rot_z1 = $mat3::from_rotation_z(deg(180.0));
             let rot_z2 = $mat3::from_axis_angle($vec3::Z, deg(180.0));
             assert_approx_eq!(rot_z1, rot_z2);
+            let rot_z3 = $mat3::from_quat($quat::from_rotation_z(deg(180.0)));
+            assert_approx_eq!(rot_z1, rot_z3);
+
+            should_glam_assert!({ $mat3::from_axis_angle($vec3::ZERO, 0.0) });
+            should_glam_assert!({ $mat3::from_quat($quat::from_xyzw(0.0, 0.0, 0.0, 0.0)) });
         }
 
         #[test]
@@ -133,6 +158,8 @@ macro_rules! impl_mat3_tests {
             assert_eq!($vec2::new(2.0, 0.0), m.transform_point2($vec2::X));
             assert_eq!($vec2::new(0.0, 4.0), m.transform_point2($vec2::Y));
 
+            should_glam_assert!({ $mat3::from_scale($vec2::ZERO) });
+
             let m = $mat3::from_scale_angle_translation(
                 $vec2::new(0.5, 1.5),
                 $t::to_radians(90.0),
@@ -184,6 +211,15 @@ macro_rules! impl_mat3_tests {
             assert_approx_eq!($newvec3(0.0, 0.0, 8.0), m.z_axis);
         }
 
+        #[test]
+        fn test_from_mat2() {
+            let m2 = $mat2::from_cols_array_2d(&[[1.0, 2.0], [3.0, 4.0]]);
+            let m3 = $mat3::from_mat2(m2);
+            assert_eq!(
+                $mat3::from_cols_array_2d(&[[1.0, 2.0, 0.0], [3.0, 4.0, 0.0], [0.0, 0.0, 1.0]]),
+                m3
+            );
+        }
         #[test]
         fn test_from_mat4() {
             let m4 = $mat4::from_cols_array_2d(&[
@@ -192,7 +228,7 @@ macro_rules! impl_mat3_tests {
                 [9.0, 10.0, 11.0, 12.0],
                 [13.0, 14.0, 15.0, 16.0],
             ]);
-            let m3 = $mat3::from(m4);
+            let m3 = $mat3::from_mat4(m4);
             assert_eq!(
                 $mat3::from_cols_array_2d(&[[1.0, 2.0, 3.0], [5.0, 6.0, 7.0], [9.0, 10.0, 11.0]]),
                 m3
@@ -253,6 +289,8 @@ macro_rules! impl_mat3_tests {
             assert_approx_eq!($mat3::IDENTITY, m * m_inv);
             assert_approx_eq!($mat3::IDENTITY, m_inv * m);
             assert_approx_eq!(m_inv, rotz_inv * scale_inv);
+
+            should_glam_assert!({ $mat3::ZERO.inverse() });
         }
 
         #[test]
@@ -269,6 +307,22 @@ macro_rules! impl_mat3_tests {
             assert_eq!($mat3::ZERO, m0 - m0);
             assert_approx_eq!(m0, m0 * $mat3::IDENTITY);
             assert_approx_eq!(m0, $mat3::IDENTITY * m0);
+
+            let mut m1 = m0;
+            m1 *= 2.0;
+            assert_eq!(m0x2, m1);
+
+            let mut m1 = m0;
+            m1 += m0;
+            assert_eq!(m0x2, m1);
+
+            let mut m1 = m0;
+            m1 -= m0;
+            assert_eq!($mat3::ZERO, m1);
+
+            let mut m1 = $mat3::IDENTITY;
+            m1 *= m0;
+            assert_approx_eq!(m0, m1);
         }
 
         #[test]
@@ -285,6 +339,9 @@ macro_rules! impl_mat3_tests {
             let mut out: [$t; 9] = Default::default();
             m.write_cols_to_slice(&mut out);
             assert_eq!(MATRIX1D, out);
+
+            should_panic!({ $mat3::from_cols_slice(&[0.0; 8]) });
+            should_panic!({ $mat3::IDENTITY.write_cols_to_slice(&mut [0.0; 8]) });
         }
 
         #[cfg(feature = "std")]
@@ -313,7 +370,9 @@ macro_rules! impl_mat3_tests {
 
 mod mat3 {
     use super::support::deg;
-    use glam::{const_mat3, mat3, vec3, vec3a, Mat3, Mat4, Vec2, Vec3, Vec3A};
+    use glam::{
+        const_mat3, mat3, swizzles::*, vec3, vec3a, Mat2, Mat3, Mat4, Quat, Vec2, Vec3, Vec3A,
+    };
 
     #[test]
     fn test_align() {
@@ -342,12 +401,14 @@ mod mat3 {
         );
     }
 
-    impl_mat3_tests!(f32, const_mat3, mat3, Mat3, Mat4, vec3, Vec3, Vec2);
+    impl_mat3_tests!(f32, const_mat3, mat3, Mat3, Mat2, Mat4, Quat, vec3, Vec3, Vec2);
 }
 
 mod mat3a {
     use super::support::deg;
-    use glam::{const_mat3a, mat3a, vec3a, Mat3A, Mat4, Vec2, Vec3, Vec3A};
+    use glam::{
+        const_mat3a, mat3a, swizzles::*, vec3a, Mat2, Mat3A, Mat4, Quat, Vec2, Vec3, Vec3A,
+    };
 
     #[test]
     fn test_align() {
@@ -372,12 +433,23 @@ mod mat3a {
         );
     }
 
-    impl_mat3_tests!(f32, const_mat3a, mat3a, Mat3A, Mat4, vec3a, Vec3, Vec2);
+    impl_mat3_tests!(
+        f32,
+        const_mat3a,
+        mat3a,
+        Mat3A,
+        Mat2,
+        Mat4,
+        Quat,
+        vec3a,
+        Vec3,
+        Vec2
+    );
 }
 
 mod dmat3 {
     use super::support::deg;
-    use glam::{const_dmat3, dmat3, dvec3, DMat3, DMat4, DVec2, DVec3};
+    use glam::{const_dmat3, dmat3, dvec3, swizzles::*, DMat2, DMat3, DMat4, DQuat, DVec2, DVec3};
 
     #[test]
     fn test_align() {
@@ -386,5 +458,16 @@ mod dmat3 {
         assert_eq!(8, mem::align_of::<DMat3>());
     }
 
-    impl_mat3_tests!(f64, const_dmat3, dmat3, DMat3, DMat4, dvec3, DVec3, DVec2);
+    impl_mat3_tests!(
+        f64,
+        const_dmat3,
+        dmat3,
+        DMat3,
+        DMat2,
+        DMat4,
+        DQuat,
+        dvec3,
+        DVec3,
+        DVec2
+    );
 }
